/****************************************************************************/
/**    GNOM - Gnom Is NOt MCuT                                             **/
/**  Copyright (C) 1997 Jochen Sarrazin, DG6VJ. All righs reserved.        **/
/**                                                                        **/
/**  Kommunikation zwischen GNOM und den dazugehrigen Kanaltreibern       **/
/****************************************************************************/

#ifndef COMM_H
#define COMM_H

#include <linux/un.h>
#include <sys/types.h>


#define COMM_VERSION 4

/****************************************************************************/
/** Struktur zur Speicherung der Daten fr die einzelnen Hardware-Treiber  **/
typedef struct _s_port_info
{
  struct _s_port_info *next;		/* Zeiger auf den nchsten Port dieses Treibers */
  char port[20];			/* Portname */
  char desc[50];			/* Beschreibung */
  int baud;				/* Geschwindigkeit */
} s_port_info;

typedef struct _s_hardware
{
  struct _s_hardware *next;		/* Zeiger auf das nchste Element der Liste */
  char name[11];			/* Name des Treibers */
  char version[5];			/* Versionsnummer */
  char autor[11];			/* Autor */
  s_port_info *ports;			/* Eine Liste der Ports auf diesen Kanal */
  int fd;
} s_hardware;
s_hardware *hardware;
/****************************************************************************/


typedef struct
{
  char port[20];
  char desc[50];
  int baud;
} s_driver_infos_port;

typedef struct
{
  char name[11];
  char version[5];
  char autor[11];
  int comm_version;
} s_driver_infos_allg;

typedef struct
{
  char port[20];		/* Portname */
  char call[10];		/* Rufzeichen */
  char path[100];		/* Pfad */
  char mycall[10];		/* eigenes Rufzeichen */
  int qsonr;			/* QSO-Nummer in diesem Treiber */
  int old;
} s_got_connect;

/*
 * Wir mchten einen Connect machen 
 */
typedef struct
{
  char port[20];
  char mycall[10];
  char call[10];
  char digis[100];
  int chan;
} s_make_connect;


/*
 * Auf diesem Port soll auf ein bestimmtes Rufzeichen gehrt werden
 */
typedef struct
{
  char port[20];
  char call[10];
} s_listen_to;


/* Defines, um 'rauszufinden, was sich gendert hat */
#define POLL_STATUS 1
#define POLL_TRIES 2
#define POLL_SQUEUE 3
#define POLL_UNACK 4



/*
 * Diverse Informationen aus dem Monitor
 * Bitte die Formate, vor allem die der Strings, 100%ig einhalten!
 *
 * Nach diesen Header-Informationen kommen die einzelnen Zeilen ohne CR dahinter.
 * Format: 2 Bytes lnge, dann der Text.
 */
#define MONI_DAMA 		1	/* Ist das Frame DAMA? */
#define MONI_AX_EAX 		2	/* AX25 oder EAX25? */
#define MONI_POLLFINAL 		4	/* Ist das Poll/Final-Bit gesetzt? */
#define MONI_BAD_HEADER 	8	/* Bad header? */

#define	MONI_PID_SEGMENT	0x08
#define	MONI_PID_ARP		0xCD
#define	MONI_PID_NETROM		0xCF
#define	MONI_PID_IP		0xCC
#define	MONI_PID_X25		0x01
#define	MONI_PID_TEXNET		0xC3
#define	MONI_PID_FLEXNET	0xCE
#define	MONI_PID_NO_L3		0xF0

#define MONI_SABM 1
#define MONI_UA 2
#define MONI_I 3
#define MONI_RR 4
#define MONI_RNR 5
#define MONI_U 6
#define MONI_DISC 7
#define MONI_DM 8
#define MONI_FRMR 9
#define MONI_UI 10
#define MONI_SABME 11
#define MONI_REJ 12

typedef struct
{
  char port[20];		/* Auf welchem Port wurde empfangen? */
  uint flags;			/* siehe MONI_* - Defines */
  char calltocall[30];		/* z.B. "DG6VJ-10->DG4IAD-12" */
  char digis[100];		/* z.B. "DB0EA DB0CPU*" */
  int frametype;		/*  1 = SABM
                                 *  2 = UA
                                 *  3 = I-Frame
                                 *  4 = RR
                                 *  5 = RNR
                                 *  6 = U
                                 *  7 = DISC
                                 *  8 = DM
                                 *  9 = FRMR
                                 * 10 = UI
                                 * 11 = SABME
                                 * 12 = REJ
                                 */
  int cmd_resp;			/* 1 = Command, 2 = Response, 0 = weder noch */
  int pid;			/* Die PID. Bitte die Zahl, einen Text macht GNOM draus. */
  unsigned char rxcount;	/* RX-Frame-Counter-Position (0-7) */
  unsigned char txcount;	/* TX-Frame-Counter-Position (0-7) */
  int anz_datalines;		/* Wieviele Datenzeilen kommen nach diesem Header? */
  char frmr_info[100];		/* Nhere Infos zu der frmr-sache */
} s_monitor;



/************************************************************************************************/


/*
 * Trgt die Informationen eines Treibers in die Treiberliste ein. Die
 * Port-Liste bleibt noch leer.
 *
 * Gibt einen Pointer auf die Hardware-Struktur zurck.
 */
s_hardware *comm_add_driver(s_driver_infos_allg infos_allg, int fd);


/*
 * Trgt diesen Port in die bergebene Hardware-Struktur ein.
 */
void comm_add_driverport(s_hardware *hard, s_driver_infos_port portinfo);


/*
 * Diese Funktion initialisiert die Treiber. Sie klappert alle sockets ab, die in
 * gnom.conf unter driver_sockets angegeben wurden, und holt sich da Konfigurations-
 * Infos 'raus.
 */
void comm_init_drivers();


/* ein Monitor-Frame kommt */
void comm_incoming_monitordata(s_hardware *hard);


/* Von Auen kommt ein connect an */
void comm_incoming_connect(s_hardware *hard);


int comm_which_channel(int driverfd, int qsonr);

void comm_incoming_disconect(s_hardware *hard);


/*
 * Hier kommt die Antwort auf einen Connect-Befehl
 */
void comm_incomming_connectanswer(s_hardware *hard);


/*
 * Ankommende Daten
 */
void comm_incoming_data(s_hardware *hard);


/*
 * Irgendwas in den /proc/net/ax25 hat sich gendert. Rausfinden was und
 * entsprechend reagieren
 */
void comm_incoming_status(s_hardware *hard);


/*
 * Alle Kanle connecten, die diesen Treiber benutzt haben
 */
void comm_disconnect_this_driver(s_hardware *hard);


/*
 * Der Treiber hat das empfangene Frame verstanden und hat noch Platz
 * fr ein weiteres
 */
void comm_incoming_send_next_frame(s_hardware *hard);


/*
 * Der Treiber kann keine weiteren Daten mehr verarbeiten. Also schicken
 * wir ihm keine mehr. (Flag auf 0 setzen)
 */
void comm_incoming_no_more_frames(s_hardware *hard);


/*
 * Dieser Treiber will was von uns
 */
void comm_driver_command(s_hardware *hard);


void comm_close_connection(int chan);


void comm_make_connect(int chan, char *call, char *path, char *mycall, char *port, int driverfd);


void comm_send_frame(int chan, char *str, int len);


/*
 * Auf dieses Rufzeichen auf diesem Port lauschen
 */
void comm_listen_to(char *call, s_port_info *port, int fd);


/*
 * Auf dieses Rufzeichen auf diesem Port nicht mehr lauschen
 */
void comm_dont_listen_to(char *call, s_port_info *port, int fd);


void comm_monitor_activate();


void comm_monitor_deactivate();


#endif

