/*
 *  GNOM - packet radio-terminal for Linux
 *  Copyright (C) 1997 Jochen Sarrazin, DG6VJ. All rights reserved.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/****************************************************************************/
/**  Hilfs-Funktionen, die auf GNOM festgelegt sind. Funktionen, die auch  **/
/**  von anderen Programmen benutzt werden knnen, sind in toolbox.h.      **/
/****************************************************************************/

#include "global.h"
#include "utils.h"

#include "toolbox.h"
#include "sndqueue.h"
#include "display.h"
#include "speaker.h"
#include "dostime.h"
#include "password.h"
#include "comm.h"
#include "crc.h"
#include "remote.h"
#include "log.h"
#include "xabin.h"

#include <string.h>

/* System-Meldungen */
static char disc_str[] = "DISCONNECTED fm";



typedef struct
{
  int chancount;
} s_status_allg;

/*typedef struct
{
} s_status_chan;*/

/*
*** get_hex: some compilers have real big trouble when reading hex values from
***          a file with fscanf() that have leading zeros! e.g. 00A will be
***          read as two separate values (0 and A)! grr!!
***          get_hex skips all leading zeros to eliminate the problem.
***
 */

uint get_hex (char *hex)
{
  register i = 0;
  uint   ret = 0;

  while (hex[i] == '0')
    i++;
  sscanf(hex+i, "%x", &ret);
  return (ret);
}



void _randomize()
{
  int t;
  
  t = (int)time(NULL);
  srand(t);
}

int _randint(short my_range)
{
  int i,j;
  
  i = RAND_MAX / my_range;
  i *= my_range;
  while ((j = rand()) >= i) continue;
  return (j % i) % my_range;
}

/* randomize erzeugt eine Zufallszahl in den Grenzen LO - HI */
/* wird fuer die Passwordberechnungen gebraucht              */

short dp_randomize(short low, short hiw)
{
  short erg;
  
  erg = _randint(hiw - low + 1);
  return (erg + low);
}



char *spec_time(int zeit)
{
  char *str,tmp[500];
  if (zeit < 100)	/* Sekunden */
     sprintf(tmp,"%i secs",zeit);
  else
  {
    zeit = zeit / 60;		/* Minuten */
    if (zeit < 99)
       sprintf(tmp,"%i mins",zeit);
    else
    {
       zeit = zeit / 60;	/* Stunden */
       sprintf(tmp,"%i hours",zeit);
    }
  }
  
  str = (char *)strdup(tmp);
  return(str);
}



/* Sucht ein cookie aus dem cookie-File und sendet es auf dem angegebenen Kanal aus. */
void send_cookie(int chan)
{
  int i,j,k,this_one=FALSE,ready=FALSE;
  FILE *fs;
  char tmp[500];
  
  
  sprintf(tmp,"%s/cookie",config->maindir);
  if ((fs = fopen(tmp,"r")) == NULL) return;
  
  /* cookies zhlen */
  i = 0;
  while (fgets(tmp,400,fs) != NULL)
  {
    if (tmp[0] == '@') i++;
  }
  i--;;
  
  /* An den Anfang des Files springen */
  fseek(fs,0,SEEK_SET);
  
  /* Zufallszahl  */
  srand(rand());
  k = 1+(int) ((float)i*rand()/(RAND_MAX+1.0));
  
  i = 0;
  while ((fgets(tmp,400,fs) != NULL) && (!ready))
  {
    if (tmp[0] == '@') {
      i++;
      if (this_one) {
        ready = TRUE;
        this_one = FALSE;
      }
      else if (i == k) this_one = TRUE;
    }
    else if (this_one)
         {
           if ((j = POS('\xA',tmp)) != -1) tmp[j] = '\0';
           strcat(tmp,"\xD");
           send_Textline(chan,tmp);
         }
  }
  
  fclose(fs);
}



/* Guckt, ob noch ein Kanal frei ist */
int chan_free(char *mycall)
{
  int i=1;
  
  while ((i <= chancount) && ((chaninfo[i]->call[0] != '\0') || (strcmp(mycall,chaninfo[i]->mycall))))
     i++;
  
  if (i > chancount)
     return(-1);
  else
     return(i);
}



void disc_channel(int chan)
{
  char tmp[100];
  s_conncalls *cc_tmp;

  /* Einen Eintrag ins Logfile machen */
  log_connection(chan);

  /* Wenn auf dieses mycall nicht gelauscht werden sollte, die lauscherei jetzt wieder aktivieren. */
  check_for_mycall_disconnect(chaninfo[chan]->mycall);

  if (chan != AktChan)
  {
    chaninfo[chan]->flags |= CH_NEWDATA;
    update_lower_statusline(TRUE);
  }
  sprintf(tmp,"\r*** (%s) %s %s%s\r",chaninfo[chan]->port,disc_str,chaninfo[chan]->firstcall,chaninfo[chan]->digis);
  display_addline(tmp,strlen(tmp),chan,ATTR_RX_STATUS,TRUE);
  
  /* Disconnect - Bimmeln */
  if (config->tones->disc != NULL)
     if (config->tones->bell) speaker_playstring(config->tones->disc);
  
  /* Dem Treiber sagen, das disconnected werden soll */
  if (chaninfo[chan]->qsonr != -1) comm_close_connection(chan);
  
  /* Speicher freigeben etc */
  if (chaninfo[chan]->name != NULL)
  {
    free(chaninfo[chan]->name);
    chaninfo[chan]->name = NULL;
  }
  if (chaninfo[chan]->digis != NULL)
  {
    free(chaninfo[chan]->digis);
    chaninfo[chan]->digis = NULL;
  }
  if (chaninfo[chan]->AutoEBIN != NULL)
  {
    free(chaninfo[chan]->AutoEBIN);
    chaninfo[chan]->AutoEBIN = NULL;
  }
  
  chaninfo[chan]->driverfd = -1;
  chaninfo[chan]->qsonr = -1;
  chaninfo[chan]->call[0] = '\0';
  chaninfo[chan]->tries = 0;
  chaninfo[chan]->unack = 0;
  chaninfo[chan]->squeue = 0;
  chaninfo[chan]->status = 0;
  
  chaninfo[chan]->rx_bytes = 0;
  chaninfo[chan]->tx_bytes = 0;
  
  /* Wenn gerade ein AutoBIN-File empfangen wird, abbrechen */
  if ((chaninfo[chan]->flags & CH_AUTOBINRX) == CH_AUTOBINRX)
  {
    close(chaninfo[chan]->autobin->fd);
    free(chaninfo[chan]->autobin->name);
    free(chaninfo[chan]->autobin);
    chaninfo[chan]->autobin = NULL;
    
    chaninfo[chan]->flags &= ~CH_AUTOBINRX;
    chaninfo[chan]->flags |= CH_LINEMODE;
    if ((chaninfo[chan]->oldflags & CH_SAVE) == CH_SAVE) chaninfo[chan]->flags |= CH_SAVE;
  }
  
  /* Wird gerade ein 7plus-File empfangen? */
  if ((chaninfo[chan]->flags & CH_7PLUS) == CH_7PLUS)
  {
    close(chaninfo[chan]->auto7plus->fd);
    free(chaninfo[chan]->auto7plus->name);
    free(chaninfo[chan]->auto7plus);
    chaninfo[chan]->auto7plus = NULL;
    
    chaninfo[chan]->flags &= ~CH_7PLUS;
  }
  
  if (chaninfo[chan]->pwinfo != NULL)
  {
    if (chaninfo[chan]->pwinfo->pw != NULL)
       free(chaninfo[chan]->pwinfo->pw);
    if (chaninfo[chan]->pwinfo->dbtime != NULL)
       free(chaninfo[chan]->pwinfo->dbtime);
    free(chaninfo[chan]->pwinfo);
    chaninfo[chan]->pwinfo = NULL;
  }
  
  /* sendqueue leer machen */
  while (chaninfo[chan]->sendqueue != NULL)
     NextQueueEntry(chan);
  
  /* conncalls leer machen */
  while (chaninfo[chan]->conncalls != NULL)
  {
    cc_tmp = chaninfo[chan]->conncalls;
    chaninfo[chan]->conncalls = chaninfo[chan]->conncalls->next;
    free(cc_tmp);
  }

  
  chaninfo[chan]->flags = CH_STDFLAGS;
  
  if (chan == AktChan) display_newscreen();
}



void macro_thisline(char *dest, char *src, int chan)
{
  int i,k,count;
  char help[500];
  struct tm sTm;
  time_t akttime;


  count = strlen(src);

  i = 0;
  k = 0;
  while (i < count)
  {
    switch(src[i])
    {
      case '\n': dest[k] = '\r';
                 k++;
                 break;
      case '%': i++;
                if (i < count)
                   switch(src[i])
                   {
                     case '%': dest[k] = '%';
                               k++;
                               break;
                     case 'V': /* Versions-Info mit Patchlevel */
                               dest[k] = '\0';
                               sprintf(help,"%i.%i.%i",MAJORLEVEL,MINORLEVEL,PATCHLEVEL);
                               memcpy(dest+k,help,strlen(help));
                               k = k + strlen(help);
                               break;
                     case 'v': /* Versions-Info ohne Patchlevel */
                               dest[k] = '\0';
                               sprintf(help,"%i.%i",MAJORLEVEL,MINORLEVEL);
                               memcpy(dest+k,help,strlen(help));
                               k = k + strlen(help);
                               break;
                     case 'k': /* Kanal, auf dem die Gegenstation hier eingeloggt ist */
                               dest[k] = '\0';
                               sprintf(help,"%i",chan);
                               memcpy(dest+k,help,strlen(help));
                               k = k + strlen(help);
                               break;
                     case 'n': /* Wenn der Name der Gegenstation bekannt ist, wird dieser
                                * ausgesendet, ansonsten das Rufzeichen.
                                */
                               dest[k] = '\0';
                               if (chaninfo[chan]->name != NULL)
                                  strcpy(help,chaninfo[chan]->name);
                               else
                                  strcpy(help,chaninfo[chan]->call);
                               memcpy(dest+k,help,strlen(help));
                               k = k + strlen(help);
                               break;
                     case 'D': /* Einlog-Datum */
                               sTm = *gmtime(&chaninfo[chan]->logintime);
                               dest[k] = '\0';
                               sprintf(help,"%.2i.%.2i.%.2i",sTm.tm_mday,sTm.tm_mon+1,sTm.tm_year);
                               memcpy(dest+k,help,strlen(help));
                               k = k + strlen(help);
                               break;
                     case 'd': /* Aktuelles Datum */
                               akttime = time(NULL);
                               sTm = *gmtime(&akttime);
                               dest[k] = '\0';
                               sprintf(help,"%.2i.%.2i.%.2i",sTm.tm_mday,sTm.tm_mon+1,sTm.tm_year);
                               memcpy(dest+k,help,strlen(help));
                               k = k + strlen(help);
                               break;
                     case 'T': /* Einlog-Zeit */
                               sTm = *gmtime(&chaninfo[chan]->logintime);
                               dest[k] = '\0';
                               sprintf(help,"%.2i:%.2i:%.2i",sTm.tm_hour,sTm.tm_min,sTm.tm_sec);
                               memcpy(dest+k,help,strlen(help));
                               k = k + strlen(help);
                               break;
                     case 't': /* aktuelle Zeit */
                               akttime = time(NULL);
                               sTm = *gmtime(&akttime);
                               dest[k] = '\0';
                               sprintf(help,"%.2i:%.2i:%.2i",sTm.tm_hour,sTm.tm_min,sTm.tm_sec);
                               memcpy(dest+k,help,strlen(help));
                               k = k + strlen(help);
                               break;
                     case 'a': /* away-Zeit */
                               sTm = *gmtime(&awaytime);
                               dest[k] = '\0';
                               sprintf(help,"%.2i:%.2i:%.2i",sTm.tm_hour,sTm.tm_min,sTm.tm_sec);
                               memcpy(dest+k,help,strlen(help));
                               k = k + strlen(help);
                               break;
                     case 'm': /* eigenes Call */
                               dest[k] = '\0';
                               strcpy(help,chaninfo[chan]->mycall);
                               memcpy(dest+k,help,strlen(help));
                               k = k + strlen(help);
                               break;
                     case 'c': /* Call der Gegenstation */
                               dest[k] = '\0';
                               strcpy(help,chaninfo[chan]->call);
                               memcpy(dest+k,help,strlen(help));
                               k = k + strlen(help);
                               break;
                     case 'p': /* Portname */
                               dest[k] = '\0';
                               strcpy(help,chaninfo[chan]->port);
                               memcpy(dest+k,help,strlen(help));
                               k = k + strlen(help);
                               break;


                     default: dest[k] = '%'; k++;
                              dest[k] = src[i]; k++;
                   }
                break;
      default: dest[k] = src[i];
               k++;
    }
    
    i++;
  }
  
  
  dest[k] = '\0';
}



/* Verschickt das bergebene File, und ersetzt die Makros */
void send_macrofile(char *filename, int chan)
{
  int fd,i,count;
  char tmp[500],str[500],dest[500],output[1000];
  
  if ((fd = open(filename,O_RDONLY)) == -1) return;
  
  while ((count = read(fd,tmp,256)) > 0)
  {
    tmp[count] = '\0';
    output[0] = '\0';
    while (tmp[0] != '\0')
    {
      /* Das Frame in einzelne Zeilen zerlegen */
      if ((i = POS('\r',tmp)) != -1)
      {
        /* Return kommt drin vor */
        COPY(str,tmp,0,i);
        COPY(tmp,tmp,i,strlen(str)+i);
      }
      else
      {
        /* Kein Return - das war das Ende */
        strcpy(str,tmp);
        tmp[0] = '\0';
      }
      
      macro_thisline(dest,str,chan);
      strcat(output,dest);
    }
    
    send_Textline(chan,output);
    output[0] = '\0';
  }
  
  close(fd);
}



void ClearUserData(s_udata *udata)
{
  udata->name[0] = '\0';
  udata->add_flags = 0;
  udata->sub_flags = 0;
  udata->pwmode = 0;
  udata->password[0] = '\0';
  udata->type = TYPE_TERMINAL;		/* Standard-Typ: Terminal */
}



/* User-Informationen wenn vorhanden aus dem User-Info-File holen und
   in die Datenstruktur schreiben. */
int GetUserData(s_udata *udata)
{
  int f;
  char filename[500],line[500],str1[500],str2[5];
  int i;
  FILE *fs;
  
  
  ClearUserData(udata);
  
  strcpy(str1,udata->call);
  Klein(str1);
  COPY(str2,str1,0,3);
  sprintf(filename,"%s/userdb/%s.../%s",config->maindir,str2,str1);
  
  
  if ((f = open(filename,O_RDONLY)) == -1) return(FALSE);
  
  fs = fdopen(f,"r");
  
  while (fgets(line,499,fs) != NULL) {
    if (line[0] != '#') {	/* keine Kommentare */
      if ((i = POS('\n',line)) != -1) line[i] = '\0';
      if ((i = POS('\r',line)) != -1) line[i] = '\0';
    
      if ((line[0] != '\0') && ((i = POS(' ',line)) != -1)) {
        COPY(str1,line,0,i);
        COPY(line,line,i+1,strlen(line)-i-1);
        
        if (!strcmp(str1,"name")) {
          if (strlen(line)>20) line[21] = '\0';
          strcpy(udata->name,line);
        }
        if (!strcmp(str1,"add_flags"))
           udata->add_flags = atoi(line);
        if (!strcmp(str1,"sub_flags"))
           udata->sub_flags = atoi(line);
        if (!strcmp(str1,"pwmode"))
           udata->pwmode = atoi(line);
        if (!strcmp(str1,"password"))
           strcpy(udata->password,line);
        if (!strcmp(str1,"type"))
        {
          udata->type = atoi(line);
          if (udata->type == 0) udata->type = 1;
        }
      }
    }
  }
  
  close(f);
  
  return(TRUE);
}



/* Userdaten in das File schreiben */
void WriteUserData(const s_udata *udata)
{
  int f;
  char filename[500],str1[256],str2[5];
  
  strcpy(str1,udata->call);
  Klein(str1);
  COPY(str2,str1,0,3);
  sprintf(filename,"%s/userdb/%s...",config->maindir,str2);
  
  /* Verzeichnis erzeugen, wenns noch nicht der da ist */
  if (mkdir(filename,448) == -1) {
  }
  sprintf(filename,"%s/%s",filename,str1);
  
  
  f = open(filename,(O_WRONLY|O_CREAT));
  
  /* Zugriffsrechte einstellen */
  fchmod(f,S_IRUSR|S_IWUSR);


  sprintf(str1,"name %s\xA",udata->name);
  write(f,str1,strlen(str1));
  sprintf(str1,"add_flags %i\xA",udata->add_flags);
  write(f,str1,strlen(str1));
  sprintf(str1,"sub_flags %i\xA",udata->sub_flags);
  write(f,str1,strlen(str1));
  sprintf(str1,"pwmode %i\xA",udata->pwmode);
  write(f,str1,strlen(str1));
  sprintf(str1,"password %s\xA",udata->password);
  write(f,str1,strlen(str1));
  sprintf(str1,"type %i\xA",udata->type);
  write(f,str1,strlen(str1));
  
  close(f);
}



/* SSID vom Call abschneiden */
void CutSSID(char *dest, const char *src)
{
  int i=0;
  while ((src[i] != '-') && (src[i] != '\0'))
  {
    dest[i] = src[i];
    i++;
  }
  dest[i] = '\0';
}



void setname(char *call, char *name)
{
  s_udata udata;
  
  strcpy(udata.call,call);
  if (!GetUserData(&udata)) ClearUserData(&udata);
  
  name[21] = '\0';
  strcpy(udata.name,name);
  
  WriteUserData(&udata);
}


void settype(char *call, int type)
{
  s_udata udata;
  
  strcpy(udata.call,call);
  if (!GetUserData(&udata)) ClearUserData(&udata);
  
  udata.type = type;
  
  WriteUserData(&udata);
}



int valid_callsign(char *call)
{
  char s_ssid[100],tmp[100];
  int i,ssid;

  strcpy(tmp,call);

  if ((i = POS('-',tmp)) == -1)
  {
    /* Keine SSID - nur auf Lnge prfen */
    if (strlen(tmp) > 6) return(FALSE);
  }
  else
  {
    /* SSID - 'rauskopieren und checken */
    COPY(s_ssid,tmp,i+1,strlen(tmp)-i-1);
    tmp[i] = '\0';
    ssid = atoi(s_ssid);
    
    if ((strlen(tmp) > 6) || (ssid > 15)) return(FALSE);
  }
  
  return(TRUE);
}



/* Holt sich die Userdaten zu dem aktuellen Rufzeichen */
void get_userdata_chan(int channel)
{
  int found=FALSE;
  s_udata udata;
  char tmp[10];

  /* Infos ber diesen User aus der Datenbank holen */
  strcpy(udata.call,chaninfo[channel]->call);
  if (GetUserData(&udata) == 0)
  {
    /* Wenns den User mit dieser SSID nicht gibt, suchen wir mal einen ohne SSID */
    strcpy(tmp,udata.call);
    CutSSID(udata.call,tmp);
    found = GetUserData(&udata);
  }
  else found = TRUE;
  
  if (found)
  {
    if (udata.name[0] != '\0')
       chaninfo[channel]->name = (char *)strdup(udata.name);
    else
       chaninfo[channel]->name = NULL;
    chaninfo[channel]->type = udata.type;
  }
  else
  {
    chaninfo[channel]->name = NULL;
    chaninfo[channel]->type = TYPE_TERMINAL;
  }
}



void data_save_frame(char *data, int len, int chan)
{
  char *tmp;
  int tmplen;
  
  if (chaninfo[chan]->rxframes == NULL)
  {
    /* Noch nix vorhanden - einfach in den String 'reinschreiben */
    chaninfo[chan]->rxframes = (s_rxframes *)malloc(sizeof(s_rxframes));
    chaninfo[chan]->rxframes->text = (char *)malloc(len+1);
    chaninfo[chan]->rxframes->len = len;
    memcpy(chaninfo[chan]->rxframes->text,data,len);
    return;
  }
  
  tmp = (char *)malloc(chaninfo[chan]->rxframes->len);
  tmplen = chaninfo[chan]->rxframes->len;
  memcpy(tmp,chaninfo[chan]->rxframes->text,chaninfo[chan]->rxframes->len);
  
  free(chaninfo[chan]->rxframes->text);
  chaninfo[chan]->rxframes->text = (char *)malloc(chaninfo[chan]->rxframes->len + len+1);
  memcpy(chaninfo[chan]->rxframes->text,tmp,chaninfo[chan]->rxframes->len);
  memcpy(chaninfo[chan]->rxframes->text+chaninfo[chan]->rxframes->len,data,len);
  chaninfo[chan]->rxframes->len = chaninfo[chan]->rxframes->len + len;
  
  free(tmp);
}


int data_next_cr(char *str, int max)
{
  int i=0;
  
  while ((str[i] != '\xD') && (i < max)) i++;
  
  if (i == max) return(-1);
  return(i);
}



void conncall_addlist(int chan, char *call)
{
  s_conncalls *tmp;
  
  if (chaninfo[chan]->conncalls == NULL)
  {
    chaninfo[chan]->conncalls = (s_conncalls *)malloc(sizeof(s_conncalls));
    tmp = chaninfo[chan]->conncalls;
  }
  else
  {
    tmp = chaninfo[chan]->conncalls;
    while (tmp->next != NULL) tmp = tmp->next;
    
    tmp->next = (s_conncalls *)malloc(sizeof(s_conncalls));
    tmp = tmp->next;
  }
  
  tmp->next = NULL;
  strcpy(tmp->call,call);
}



/* Guckt, ob dieses Rufzeichen in der Liste enthalten ist. Wenn ja werden alle Calls
 * ab diesem Rufzeichen aus der Liste gelscht.
 */
int conncall_checkcall(int chan, char *call)
{
  s_conncalls *tmp,*last;

  /* Gucken, ob wir zum allerersten Call zurck fallen */
  if (!strcmp(call,chaninfo[chan]->firstcall))
  {
    /* Jau - alle Eintrge in conncalls lschen */
    tmp = chaninfo[chan]->conncalls;
    
    while (tmp != NULL)
    {
      last = tmp;
      tmp = tmp->next;
      free(last);
    }
    chaninfo[chan]->conncalls = NULL;
    
    return(TRUE);
  }
  
  
  tmp = chaninfo[chan]->conncalls;
  last = NULL;
  
  while (tmp != NULL)
  {
    if (!strcmp(call,tmp->call))
    {
      /* Call gefunden. Alles ab diesem (incl.) free'en */
      if (last != NULL) last->next = NULL;
      while (tmp != NULL)
      {
        last = tmp;
        tmp = tmp->next;
        free(last);
      }
      
      return(TRUE);
    }
    last = tmp;
    tmp = tmp->next;
  }
  
  return(FALSE);
}



void check_autobin(char *str, int chan)
{
  int i,fd,ready;
  uint len=-1, crc=-1;
  char tmp[500],name[500]="";
  unsigned long dosdatetime;
  time_t unixtime;
  struct utimbuf utimetmp;
  
  
  COPY(str,str,5,strlen(str)-5);
  
  /* Die Lnge des BIN-Teils - *MUSS* kommen */
  if ((i = POS('#',str)) != -1)
  {
    COPY(tmp,str,0,i);
    COPY(str,str,i+1,strlen(str)-i-1);
  }
  else
  {
    strcpy(tmp,str);
    str[0] = '\0';
  }
  len = atoi(tmp);
  
  if (len == 0)
  {
    send_Textline(chan,"#NO# - invalid filelength\xD");
    return;
  }
  
  ready = FALSE;
  while (!ready)
  {
    if ((i = POS('#',str)) == -1)
    {
      ready = TRUE;
      strcpy(tmp,str);
      str[0] = '\0';
    }
    else {
      COPY(tmp,str,0,i);
      COPY(str,str,i+1,strlen(str)-i-1);
    }
    
    switch (tmp[0])
    {
      case '|': /* CRC */
                crc = atoi(tmp+1);
                break;
      case '$': /* Datum/Uhrzeit-Code */
                dosdatetime = atoi(tmp+1);
                break;
      default: /* Name */
               strcpy(name,tmp);
               break;
    }
  }
  
  
  if (name[0] == '\0')
  {
    /* Kein Name vorhanden */
    i = 0;
    sprintf(name,"%s/abin/autobin.%i",config->maindir,i);
    while (file_exist(name)==1)
    {
      i++;
      sprintf(name,"%s/abin/autobin.%i",config->maindir,i);
    }
  }
  /* Ein Dateiname ist vorhanden - einen evtl. vorhandenen Pfad lschen */
  while ((i = POS('/',name)) != -1) COPY(name,name,i+1,strlen(name)-i-1);
  while ((i = POS('\\',name)) != -1) COPY(name,name,i+1,strlen(name)-i-1);
  if ((i = POS(':',name)) != -1) COPY(name,name,i+1,strlen(name)-i-1);
  
  sprintf(tmp,"%s/abin/%s",config->maindir,name);
  
  if ((fd = open(tmp,O_WRONLY|O_CREAT|O_TRUNC)) == -1)
  {
    send_Textline(chan,"#NO# - cannot open destination-file\xD");
    return;
  }

  /* Zugriffsrechte einstellen */
  fchmod(fd,S_IRUSR|S_IWUSR);
  
  /* Datum + Uhrzeit einstellen */
  unixtime = date_dos2unix((dosdatetime & 0xFFFF), (dosdatetime >> 16));
  utimetmp.actime = unixtime;
  utimetmp.modtime = unixtime;
  utime(tmp,&utimetmp);
  
  /* Jetzt chaninfo[]->autobin allokieren und fllen */
  chaninfo[chan]->autobin = (s_autobin *)malloc(sizeof(s_autobin));
  chaninfo[chan]->autobin->len = len;
  chaninfo[chan]->autobin->bytes_got = 0;
  if (crc == -1)
  {
    chaninfo[chan]->autobin->crc_rx = 0;
    chaninfo[chan]->autobin->crc_tx = 0;
    chaninfo[chan]->autobin->calc_crc = FALSE;
  }
  else
  {
    chaninfo[chan]->autobin->crc_rx = crc;
    chaninfo[chan]->autobin->crc_tx = 0;
    chaninfo[chan]->autobin->calc_crc = TRUE;
  }
  chaninfo[chan]->autobin->fd = fd;
  chaninfo[chan]->autobin->name = (char *)strdup(name);
  chaninfo[chan]->flags &= ~CH_LINEMODE;
  chaninfo[chan]->autobin->type = 1;
  
  chaninfo[chan]->flags |= CH_AUTOBINRX;
  if ((chaninfo[chan]->flags & CH_SAVE) == CH_SAVE)
  {
    chaninfo[chan]->oldflags |= CH_SAVE;
    chaninfo[chan]->flags &= ~CH_SAVE;
  }
  if (chan == AktChan) update_upper_statusline(TRUE);
  
  send_Textline(chan,"#OK#\xD");
  
  chaninfo[chan]->autobin->starttime = time(NULL);
}



/* Berechnet die CRC aus einem ganzen String */
unsigned short calc_crcthp_str(const char *str, const int len, unsigned short crc)
{
  int i;
  
  i = 0;
  while (i < len)
  {
    crcthp(str[i],&crc);
    i++;
  }
  
  return(crc);
}



void check_7plus_autosave(int chan, char *data, int len)
{
  char tmp[500],name[20];
  int part,of,i,fd,filelen=-1;
  int binbytes,blocklines;
  
  if (len < 60) return;
  
  COPY(tmp,data,0,8);
  
  if (!strcmp(tmp," go_7+. "))
  {
    /* Knnte ein 7plus-File sein */
    
    /* Steht " of " an der richtigen Stelle? */
    COPY(tmp,data,11,4);
    if (strcmp(tmp," of ")) return;
    
    /* Teile-Nummern extrahieren */
    COPY(tmp,data,8,3);
    part = atoi(tmp);
    COPY(tmp,data,16,3);
    of = atoi(tmp);
    /* Dateinamen */
    COPY(tmp,data,19,12);
    if ((i = POS('.',tmp)) != -1) tmp[i] = '\0';
    KillSpacesLeft(tmp);
    KillSpacesRight(tmp);
    Klein(tmp);
    strcpy(name,tmp);
    /* Lnge des dekodierten Files */
    COPY(tmp,data,32,7);
    binbytes = atoi(tmp);
    /* Anzahl Zeilen in einem Block */
    COPY(tmp,data,45,3);
    blocklines = get_hex (tmp);
    
    /* Lnge errechen. Hochgradig abhngig von der Teil-Nummer */
    if (part == 1)
       if (data[65] == '*')
          filelen = (blocklines+3)*70;
       else
          filelen = (blocklines+2)*70;
    else if (part == of)
            filelen = (int) ((((binbytes + 61) / 62) % blocklines)+2)*70;
         else
            filelen = (blocklines+2)*70;
    
    if ((of == 0) || (part == 0)) return;
    if (part > of) return;
    
    if (of == 1)
       sprintf(tmp,"%s/7plus/%s.7pl",config->maindir,name);
    else
       sprintf(tmp,"%s/7plus/%s.p%02x",config->maindir,name,part);
    if ((fd = open(tmp,O_WRONLY|O_CREAT)) == -1) return;
    
    /* Zugriffsrechte einstellen */
    fchmod(fd,S_IRUSR|S_IWUSR);
    
    /* Das ganze scheint ja wohl ein 7plus-File zu sein... */
    chaninfo[chan]->auto7plus = (s_auto7plus *)malloc(sizeof(s_auto7plus));
    chaninfo[chan]->auto7plus->fd = fd;
    chaninfo[chan]->auto7plus->name = (char *)strdup(tmp);
    chaninfo[chan]->auto7plus->len = filelen;
    chaninfo[chan]->auto7plus->bytes_got = len+1;
    chaninfo[chan]->auto7plus->starttime = time(NULL);
    
    chaninfo[chan]->flags |= CH_7PLUS;
    
    /* Den String schreiben */
    write(fd,data,len);
    write(fd,"\n",1);
    
    update_upper_statusline(TRUE);
    update_lower_statusline(TRUE);
  }
  else return;
}



void check_7plus_saveit(int chan, char *data, int len)
{
  char tmp[500];
  time_t timediff;
  int i;
  char *timeptr;

  /* String abspeichern */
  write(chaninfo[chan]->auto7plus->fd,data,len);
  write(chaninfo[chan]->auto7plus->fd,"\n",1);
  chaninfo[chan]->auto7plus->bytes_got += len+1;	/* das +1 ist das \n dahinter */
  update_lower_statusline(TRUE);

  COPY(tmp,data,0,10);
  if (!strcmp(tmp," stop_7+. "))
  {
    /* Das wars, das 7plus-File ist zu Ende */
    close(chaninfo[chan]->auto7plus->fd);
    while ((i = POS('/',chaninfo[chan]->auto7plus->name)) != -1)
        COPY(chaninfo[chan]->auto7plus->name,
             chaninfo[chan]->auto7plus->name,i+1,
             strlen(chaninfo[chan]->auto7plus->name)-i-1); 
    
    /* Informationen an die Gegenstation verschicken */
    timediff = time(NULL)-chaninfo[chan]->auto7plus->starttime;
    if (timediff == 0) timediff = 1;
    timeptr = (char *)spec_time(timediff);
    sprintf(tmp,"\r<GNOM>: 7plus-Autosave: '%s' saved.\r"
                "        %i/%i bytes, %s, %li baud.\r\r"
                ,chaninfo[chan]->auto7plus->name
                ,chaninfo[chan]->auto7plus->bytes_got
                ,chaninfo[chan]->auto7plus->len
                ,timeptr,(chaninfo[chan]->auto7plus->bytes_got*8)/timediff);
    free(timeptr);
    
    if (chaninfo[chan]->type == TYPE_TERMINAL)
       send_Textline(chan,tmp);
    else
       display_addline(tmp,strlen(tmp),chan,ATTR_RX_SEND,TRUE);
    
    free(chaninfo[chan]->auto7plus->name);
    free(chaninfo[chan]->auto7plus);
    chaninfo[chan]->auto7plus = NULL;
    
    chaninfo[chan]->flags &= ~CH_7PLUS;
    update_lower_statusline(TRUE);
    update_upper_statusline(TRUE);
  }
}



/*
 * AutoBIN-Save ist an.
 * Speichern also
 */
void frame_save_autobin(char *data, int len, int chan)
{
  int i,timediff;
  char *tmp,str[100];
  
  COPY(str,data,0,7);
  if (!strcmp(str,"#ABORT#"))
  {
    /* Abbruch der AutoBIN-Geschichte */
    
    /* Speicher freimachen etc */
    close(chaninfo[chan]->autobin->fd);
    free(chaninfo[chan]->autobin->name);
    free(chaninfo[chan]->autobin);
    chaninfo[chan]->autobin = NULL;
    
    chaninfo[chan]->flags &= ~CH_AUTOBINRX;
    chaninfo[chan]->flags |= CH_LINEMODE;
    update_lower_statusline(TRUE);
    if ((chaninfo[chan]->oldflags & CH_SAVE) == CH_SAVE) chaninfo[chan]->flags |= CH_SAVE;
    
    /* Nachricht schreiben */
    strcpy(str,"\r<GNOM>: AutoBIN-RX aborted by sender.\r");
    if (chaninfo[chan]->type == TYPE_TERMINAL)
       send_Textline(chan,str);
    else
       display_addline(str,strlen(str),chan,ATTR_RX_SEND,TRUE);
    return;
  }
  
  i = chaninfo[chan]->autobin->len - chaninfo[chan]->autobin->bytes_got;
  if (i > len) i = len;
  chaninfo[chan]->autobin->crc_tx = calc_crcthp_str(data,i,chaninfo[chan]->autobin->crc_tx);
  write(chaninfo[chan]->autobin->fd,data,i);
  chaninfo[chan]->autobin->bytes_got = chaninfo[chan]->autobin->bytes_got + i;
  update_lower_statusline(TRUE);
  
  /* Wenn nach dem AutoBIN-Paket noch was kam, wird das Frame ganz oben in die rxframes-liste
   * eingetragen.
   */
  if (len > i)
  {
    memmove(data,data+i,len-i);
    len = len-i;
    
    if (chaninfo[chan]->rxframes != NULL)
    {
      tmp = chaninfo[chan]->rxframes->text;
      memcpy(tmp,chaninfo[chan]->rxframes->text,chaninfo[chan]->rxframes->len);
      
      chaninfo[chan]->rxframes->text = (char *)malloc(chaninfo[chan]->rxframes->len+len);
      memcpy(chaninfo[chan]->rxframes->text,data,i);
      memcpy(chaninfo[chan]->rxframes->text+i,tmp,chaninfo[chan]->rxframes->len);
      chaninfo[chan]->rxframes->len = chaninfo[chan]->rxframes->len + len;
    }
    else
    {
      chaninfo[chan]->rxframes = (s_rxframes *)malloc(sizeof(s_rxframes));
      chaninfo[chan]->rxframes->text = (char *)malloc(len);
      memcpy(chaninfo[chan]->rxframes->text,data,len);
      
      chaninfo[chan]->rxframes->len = len;
    }
  }
  
  if (chaninfo[chan]->autobin->bytes_got == chaninfo[chan]->autobin->len)
  {
    timediff = time(NULL)-chaninfo[chan]->autobin->starttime;
    if(timediff == 0) timediff = 1;
    tmp = spec_time(timediff);

    /* Alles empfangen! Gucken, ob die Checksummen bereinstimmen */
    if ((chaninfo[chan]->autobin->crc_rx == chaninfo[chan]->autobin->crc_tx)
         || (chaninfo[chan]->autobin->calc_crc == FALSE))
       if (chaninfo[chan]->autobin->calc_crc == FALSE)
          sprintf(str,"<GNOM>: AutoBIN-RX OK. No CRC-check (time: %s, %i baud)\xD",
                       tmp,(chaninfo[chan]->autobin->len*8)/timediff);
       else
          sprintf(str,"<GNOM>: AutoBIN-RX OK. CRC: %i (time: %s, %i baud)\xD",
                      chaninfo[chan]->autobin->crc_tx,
                      tmp,(chaninfo[chan]->autobin->len*8)/timediff);
    else
       sprintf(str,"<GNOM>: AutoBIN-RX failed (CRC-ERROR %i != %i)",
       chaninfo[chan]->autobin->crc_rx,chaninfo[chan]->autobin->crc_tx);
    
    free(tmp);
    
    if (chaninfo[chan]->type == TYPE_TERMINAL)
       send_Textline(chan,str);
    else
       display_addline(str,strlen(str),chan,ATTR_RX_SEND,TRUE);
    
    /* Speicher freimachen etc */
    close(chaninfo[chan]->autobin->fd);
    free(chaninfo[chan]->autobin->name);
    free(chaninfo[chan]->autobin);
    chaninfo[chan]->autobin = NULL;
    
    chaninfo[chan]->flags &= ~CH_AUTOBINRX;
    chaninfo[chan]->flags |= CH_LINEMODE;
    update_lower_statusline(TRUE);
    if ((chaninfo[chan]->oldflags & CH_SAVE) == CH_SAVE) chaninfo[chan]->flags |= CH_SAVE;
  }
}



/* Auf diesem Kanal wurde disconnected. Wenn noch ein 7plus- oder ein AutoBIN-File
 * empfangen wird, wird der empfang abgebrochen, damit auf dem Bildschirm wieder
 * etwas angezeigt wird.
 */
void close_autobin_7plus(int chan)
{

  /* Wenn gerade ein AutoBIN-File empfangen wird, abbrechen */
  if ((chaninfo[chan]->flags & CH_AUTOBINRX) == CH_AUTOBINRX)
  {
    close(chaninfo[chan]->autobin->fd);
    free(chaninfo[chan]->autobin->name);
    free(chaninfo[chan]->autobin);
    chaninfo[chan]->autobin = NULL;
    
    chaninfo[chan]->flags &= ~CH_AUTOBINRX;
    chaninfo[chan]->flags |= CH_LINEMODE;
    if ((chaninfo[chan]->oldflags & CH_SAVE) == CH_SAVE) chaninfo[chan]->flags |= CH_SAVE;
  }

  /* Wird gerade ein 7plus-File empfangen? */
  if ((chaninfo[chan]->flags & CH_7PLUS) == CH_7PLUS)
  {
    close(chaninfo[chan]->auto7plus->fd);
    free(chaninfo[chan]->auto7plus->name);
    free(chaninfo[chan]->auto7plus);
    chaninfo[chan]->auto7plus = NULL;
    
    chaninfo[chan]->flags &= ~CH_7PLUS;
  }
  
}



void data_split_frames(int chan)
{
  char *data,*tmp,str[100];
  int len,i;



  while (chaninfo[chan]->rxframes != NULL)
  {
    /* linemode oder nicht? */
    if ((chaninfo[chan]->flags & CH_LINEMODE) == CH_LINEMODE)
    {
      tmp = chaninfo[chan]->rxframes->text;
      i = data_next_cr(tmp,chaninfo[chan]->rxframes->len);
      if (i == -1) return;
      
      if (chaninfo[chan]->rxframes->len-i == 0)
      {
        free(chaninfo[chan]->rxframes);
        chaninfo[chan]->rxframes = NULL;
      }
      else
      {
        chaninfo[chan]->rxframes->text = (char *)malloc(chaninfo[chan]->rxframes->len-i-1);
        memcpy(chaninfo[chan]->rxframes->text,tmp+i+1,chaninfo[chan]->rxframes->len-i-1);
        chaninfo[chan]->rxframes->len = chaninfo[chan]->rxframes->len - i-1;
      }
      data = (char *)malloc(i+1);
      len = i;
      memcpy(data,tmp,i);
      free(tmp);
    }
    else
    {
      /* "Frame-Mode" */
      data = chaninfo[chan]->rxframes->text;
      len = chaninfo[chan]->rxframes->len;
      free(chaninfo[chan]->rxframes);
      chaninfo[chan]->rxframes = NULL;
    }
    data[len] = '\0';
    
    /* Wird gerade etwas gespeichert? */
    if ((chaninfo[chan]->flags & CH_SAVE) == CH_SAVE)
    {
      if (chaninfo[chan]->saveplain->local == 2)
      {
        /* Nicht vom SysOp selbst aufgemacht - gucken, ob das File geschlossen werden soll */
        tmp = (char *)strdup(data);
        Gross(tmp);
        if (!strcmp(data,tmp) || (data[0] == 3)) /* 3 = CTRL-Z */
        {
          sprintf(str,"<GNOM>: File closed (%i Bytes).\r",
                      (int)lseek(chaninfo[chan]->saveplain->fd,0,SEEK_CUR));
          send_Textline(chan,str);
          /* Ja, schlieen */
          close(chaninfo[chan]->saveplain->fd);
          free(chaninfo[chan]->saveplain->filename);
          free(chaninfo[chan]->saveplain);
          chaninfo[chan]->saveplain = NULL;
          
          chaninfo[chan]->flags &= ~CH_SAVE;
          update_upper_statusline(TRUE);
        }
        free(tmp);
      }
      
      if (chaninfo[chan]->saveplain != NULL)
      {
        write(chaninfo[chan]->saveplain->fd,data,len);
        write(chaninfo[chan]->saveplain->fd,"\xA",1);
      }
    }
    
    /* Gucken, ob connected to oder reconnected to kam (wenn es freigeschaltet ist) */
    if (((chaninfo[chan]->flags & CH_CHECKCONN) == CH_CHECKCONN) &&
        ((chaninfo[chan]->type != TYPE_MAILBOX) && (chaninfo[chan]->type != TYPE_CONVERS)))
    {
      tmp = (char *)malloc(len+1);
      COPY(tmp,data,0,16);
      if (!strcmp(tmp,"*** connected to"))
      {
        /* Ja, connected to */
        COPY(tmp,data,17,strlen(data));
        Gross(tmp);
        KillSpacesLeft(tmp);
        KillSpacesRight(tmp);
        
        if (valid_callsign(tmp))
        {
          strcpy(chaninfo[chan]->call,tmp);
          get_userdata_chan(chan);
          todo_on_connect(chan);
          update_lower_statusline(TRUE);
          if (chan == AktChan) update_upper_statusline(TRUE);
          
          conncall_addlist(chan,chaninfo[chan]->call);
          
          /* Alte Passwort-Infos lschen und neue 'rein machen */
          pw_free(chan);
          pw_getinfo(chan);
        }
      }
      free(tmp);
    }
    if ((chaninfo[chan]->flags & CH_CHECKDISC) == CH_CHECKDISC)
    {
      tmp = (char *)malloc(len+1);
      COPY(tmp,data,0,18);
      if (!strcmp(tmp,"*** reconnected to"))
      {
        /* Ja, connected to */
        COPY(tmp,data,19,strlen(data));
        Gross(tmp);
        KillSpacesLeft(tmp);
        KillSpacesRight(tmp);
        
        if (valid_callsign(tmp))
        {
          if (conncall_checkcall(chan,tmp))
          {
            strcpy(chaninfo[chan]->call,tmp);
            get_userdata_chan(chan);
            todo_on_connect(chan);
            update_lower_statusline(TRUE);
            
            chaninfo[chan]->flags &= ~CH_COMP;
            
            if (chan == AktChan) update_upper_statusline(TRUE);
            /* Alte Passwort-Infos lschen und neue 'rein machen */
            pw_free(chan);
            pw_getinfo(chan);
            
            /* AutoBIN- und 7plus-Speicherung beenden und wieder auf linemode schalten */
            if (((chaninfo[chan]->flags & CH_7PLUS) != 0) ||
                ((chaninfo[chan]->flags & CH_AUTOBINRX) != 0))
            {
              close_autobin_7plus(chan);
              chaninfo[chan]->flags |= CH_LINEMODE;
              display_addline("\xD",1,chan,ATTR_RX_EMPF,FALSE);
              display_addline(data,len,chan,ATTR_RX_EMPF,TRUE);
            }
          }
        }
      }
      free(tmp);
    }
    
    
    /* Wenn kein AutoBIN am Laufen ist, wird geguckt, ob ein #BIN#-Header anfngt. */    
    if (((chaninfo[chan]->flags & CH_AUTOBINRX) == 0) &&
       ((chaninfo[chan]->flags & CH_CHECKAUTOBIN) == CH_CHECKAUTOBIN))
    {
      if (!strncmp("#EBIN#",data,6)) ebin_check_ebinline(data,chan,len);
      if (!strncmp("#BIN#",data,5)) check_autobin(data,chan);
    }
    else if ((chaninfo[chan]->flags & CH_AUTOBINRX) != 0)
            if (chaninfo[chan]->autobin->type == 1)
               frame_save_autobin(data,len,chan);
            else
               ebin_saveframe(data,len,chan);
    
    if ((chaninfo[chan]->flags & CH_CHECK7PLUS) == CH_CHECK7PLUS)
    {
      /* 7plus-Autosave? */    
      if ((chaninfo[chan]->flags & CH_7PLUS) == 0)
         check_7plus_autosave(chan,data,len);
      else
         check_7plus_saveit(chan,data,len);
    }
    
    
    if (chaninfo[chan]->pwinfo != NULL)
       if (chaninfo[chan]->pwinfo->pwmode == 2)		/* BayBox-Passwort? */
          if (chaninfo[chan]->pwinfo->dbtime == NULL)	/* Noch keine Zeit gespeichert? */
              pw_check_baycom(chan,data);	/* Jau! */
    
    
    /* Wenn wir ein DieBox-Passwort haben, brauchen wir das Datum. Also gucken */
    if (chaninfo[chan]->pwinfo != NULL)
       if ((chaninfo[chan]->pwinfo->pwmode == 4) || (chaninfo[chan]->pwinfo->pwmode == 5))
           pw_check_for_dieboxtime(chan,data);
    
    
    switch (chaninfo[chan]->consstat)
    {
      case 1: /* AutoBIN-Zeile abgeschickt, warte auf Antwort */
              if (!strncmp(data,"#OK#",4))
              {
                send_AutoBIN(chan);
                chaninfo[chan]->consstat = 0;
              }
              else if (!strncmp(data,"#EOK#",5))
                   {
                     strcpy(str,data+5);
                     chaninfo[chan]->AutoEBINoffset = atoi(str);
                     if (chaninfo[chan]->AutoEBINoffset == -1) chaninfo[chan]->AutoEBINoffset = 0;
                     send_AutoEBIN(chan);
                     chaninfo[chan]->consstat = 4;
                   }
                   else if ((!strncmp(data,"#NO#",4)) || (!strncmp(data,"#ABORT#",7)))
                        {
                          chaninfo[chan]->consstat = 0;
                          
                          /* chaninfo[chan]->autobin wieder frei machen */
                          if (chaninfo[chan]->autobin->name != NULL) free(chaninfo[chan]->autobin->name);
                          free(chaninfo[chan]->autobin);
                          chaninfo[chan]->autobin = NULL;
                        }
                        break;
      case 2: /* RMNC-Passwort-Abfrage */
              pw_check_rmncpw(chan,data);
              break;
      case 3: /* MD2-Passwort-Abfrage */
              pw_check_md2pw(chan,data);
              break;
      case 4: /* EBIN-bertragung */
              ebin_commands(chan,data,len);
              break;
      case 5: /* BayBox-Sysop-Passwort-Abfrage */
              pw_check_baycom(chan,data);
              break;
    }
    
    
    /* Remote-Befehl? */
    if (((data[0] == '/') && (data[1] == '/') && ((chaninfo[chan]->flags & CH_REMOTE) == CH_REMOTE))
       || ((!strncmp(data,"//COMP",6)) && (chaninfo[chan]->type == TYPE_MAILBOX)))
    {
      memmove(data,data+2,len-2);
      len = len - 2;
      remote_command(data,len,chan);
    }
    
    
    free(data);
  }
}



unsigned short calc_crctph_wholefile(char *path)
{
  unsigned short crc=0;
  int fd,count,ready=FALSE;
  char tmp[500];
  
  
  if ((fd = open(path,O_RDONLY)) == -1) return(0);
  
  while (!ready)
  {
    count = read(fd,tmp,499);
    if (count != 499) ready = TRUE;
    
    crc = calc_crcthp_str(tmp,count,crc);
  }
  return(crc);
}



/* Bei jedem neuen Connect werden diese Sachen abgearbeitet */
void todo_on_connect(int chan)
{
  /* Passwort-Daten laden, wenn vorhanden */
  pw_getinfo(chan);

  switch (chaninfo[chan]->type)
  {
    case TYPE_MAILBOX:
         chaninfo[chan]->flags &= ~CH_REMOTE;
         break;
    case TYPE_DIGI:
         chaninfo[chan]->flags &= ~CH_REMOTE;
         break;
    case TYPE_CONVERS:
         chaninfo[chan]->flags &= ~CH_REMOTE;
         break;
    case TYPE_TERMINAL:
         chaninfo[chan]->flags |= CH_REMOTE;
         break;
  }
  
  
}



void write_to_chtype(char *str, chtype *chtype_str, int attrib)
{
  int i=0,len;
  
  len = strlen(str);
  
  while (i < len)
  {
    chtype_str[i] = (chtype)str[i]|attrib;
    i++;
  }
}



void chtype_write_attribs(chtype *attrarray, int start, int count, int attrib)
{
  int i;

  i = start;
  while (i < start+count)
  {
    attrarray[i] = (chtype)attrib;
    i++;
  }
}



/*
 * Wandelt nicht darstellbare Zeichen so um, da sie Angezeigt werden knnen (BIN oder so)
 *
 * Genauer werden folgende Sachen gemacht:
 *   - '\0'en im String zu ' ' umgewandelt
 *   - '\xFF'en im String zu ' ' umgewandelt
 *   - '\xA'en werden zu '\xD' umgewandelt
 */
void checkshowchars(char *tmp, int len)
{
  int i;
  
  i = 0;
  while (i < len)
  {
    switch (tmp[i])
    {
      case '\0': tmp[i] = ' '; break;
      case '\x0A': tmp[i] = '\x0D'; break;
      case '\xFF': tmp[i] = ' '; break;
      case (char)12: tmp[i] = '.'; break;
      case (char)27: tmp[i] = '.'; break;
      case (char)155: tmp[i] = '.'; break;
      
    }
    if (((uchar)tmp[i] < 32) && (tmp[i] != 0) && (tmp[i] != 13)) tmp[i] = '.';
    
    i++;
  }
}


int ax25_get_driverfd(char *portstr)
{
  s_hardware *hard;
  s_port_info *port;
  
  hard = hardware;
  while (hard != NULL)
  {
    port = hard->ports;
    while (port != NULL)
    {
      if (!strcmp(portstr,port->port)) return(hard->fd);
      port = port->next;
    }
    hard = hard->next;
  }
  
  return(-1);
}



void delete_this_driver_from_list(s_hardware *hard)
{
  s_port_info *ports,*ptmp;
  s_hardware *tmp,*last;

  if (hard == hardware)
  {
    /* Es ist der erste Eintrag */
    hardware = hardware->next;
    ports = hard->ports;
    free(hard);
  }
  else
  {
    tmp = hardware;
    while ((tmp != hard) && (tmp != NULL))
    {
      last = tmp;
      tmp = tmp->next;
    }
    if (tmp != NULL)
    {
      /* Den Eintrag hama doch noch gefunden */
      last->next = tmp->next;	/* tmp berspringen */
      ports = tmp->ports;
      free(tmp);
    }
  }
  
  while (ports != NULL)
  {
    ptmp = ports;
    ports = ports->next;
    free(ptmp);
  }
}



int valid_portname(char *name)
{
  s_hardware *hard;
  s_port_info *ports;
  
  hard = hardware;
  while (hard != NULL)
  {
    ports = hard->ports;
    while (ports != NULL)
    {
      if (!strcmp(name,ports->port))
         return(TRUE);
      ports = ports->next;
    }
    hard = hard->next;
  }
  
  return(FALSE);
}



void mycall_addcall(char *call)
{
  s_hardware *hard;
  s_port_info *port;
  
  hard = hardware;
  while (hard != NULL)
  {
    port = hard->ports;
    while (port != NULL)
    {
      comm_listen_to(call,port,hard->fd);
      port = port->next;
    }
    hard = hard->next;
  }
}



/* Diese Funktion guckt, ob das angegebene Mycall auf einem anderen nicht connecteten Port
 * eingestellt ist. Wenn nicht wird der Treiber angewiesen, auf dieses Rufzeiche nicht mehr
 * zu lauschen.
 */
void check_for_mycall_connect(char *mycall)
{
  int i=1,found=FALSE;
  s_hardware *hard;
  s_port_info *ports;
  
  while ((i <= chancount) && !found)
  {
    if (chaninfo[i]->driverfd == -1)
       if (!strcmp(mycall,chaninfo[i]->mycall)) found = TRUE;
    i++;
  }
  
  if (!found)
  {
    /* Das Rufzeichen ist nicht eingestellt - dem Treiber sagen, er soll nicht mehr drauf hren */
    hard = hardware;
    while (hard != NULL)
    {
      ports = hard->ports;
      while (ports != NULL)
      {
        comm_dont_listen_to(mycall,ports,hard->fd);
        ports = ports->next;
      }
      hard = hard->next;
    }
  }
}



/* Wenn auf das angegebene mycall nicht gelauscht werden sollte, dem Treiber bescheid
 * sagen, da doch gelauscht werden soll, weil ja der kanal mit diesem Mycall wieder
 * frei wird.
 */
void check_for_mycall_disconnect(char *mycall)
{
  int i=1,found=FALSE;
  s_hardware *hard;
  s_port_info *ports;
  
  while ((i <= chancount) && !found)
  {
    if (chaninfo[i]->driverfd == -1)
       if (!strcmp(mycall,chaninfo[i]->mycall)) found = TRUE;
    i++;
  }
  
  if (!found)
  {
    /* Das Rufzeichen ist nicht eingestellt - dem Treiber sagen, er soll nicht mehr drauf hren */
    hard = hardware;
    while (hard != NULL)
    {
      ports = hard->ports;
      while (ports != NULL)
      {
        comm_listen_to(mycall,ports,hard->fd);
        ports = ports->next;
      }
      hard = hard->next;
    }
  }
}



/* Speichert den Status der Verbindungen in ein File */
void save_status()
{
  char filename[500];
  
  
  sprintf(filename,"%s/status.gnom",config->maindir);
}



