/****************************************************************************/
/**    Kernel-AX25-Treiber fr GNOM.                                       **/
/**  Copyright (C) 1997 Jochen Sarrazin, DG6VJ. All righs reserved.        **/
/**                                                                        **/
/**  Hauptprogramm                                                         **/
/****************************************************************************/


/*
##########################################################################
##                    ATTENTIONE ALL READERS                            ##
##                                                                      ##
##  Das hier ist die erste Implementation des Protokolls. Es ist        ##
##  mit Sicherheit noch nicht alles glatt und ohne Kanten, aber es      ##
##  **LUFT**. Sollte jemandem noch etwas auffallen, bitte Mail. Es     ##
##  fehlen auch noch ein paar Befehle, die aber beim Kernel-AX25 nicht  ##
##  so ins Gewicht fallen. Anfragen bitte an mich. Ebenso knnen noch   ##
##  weitere Fehlercodes beim connect-Befehl definiert werden, die ich   ##
##  ebenfalls gerne in GNOM aufnehme.                                   ##
##########################################################################
##  If you plan to write your own driver please mail me! There will     ##
##  be minor or larger protocol-changes. I'll send such informations    ##
##  to all registered users.                                            ##
##########################################################################
*/


#include "global.h"
#include <errno.h>


/* Wird automatisch aufgerufen, wenn das Hauptprogramm beendet wird.
 * Um es ganz korrekt zu machen, werde ich hier spter den ganzen Speicher
 * freigeben, den das Programm belegt...
 */
static void MainTerminate()
{
  _exit(0);
}


/*
 * Berechnet den Wert fr maxfd neu
 */
int getmaxfd()
{
  int i,k;
  s_portinfo *tmp;
  s_listen *listento;
  
  i = sockfd;
  
  if (gnomfd > i) i = gnomfd;
  
  if (monifd > i) i = monifd;
  
  tmp = portsroot;
  while (tmp != NULL)
  {
    if (tmp->fd > i) i = tmp->fd;
    listento = tmp->listen;
    while (listento != NULL)
    {
      if (listento->fd > i) i = listento->fd;
      listento = listento->next;
    }
    tmp = tmp->next;
  }
  
  k = 0;
  while (k < MAX_QSOs)
  {
    if (qso[k] != NULL)
       if (qso[k]->fd > i) i = qso[k]->fd;
    k++;
  }
 
  return(i+2);
}



/*
 * Diverse globale Variablen initialisieren
 */
void init_globals()
{
  strcpy(driver_infos_overall.name,DRIVER_NAME);
  strcpy(driver_infos_overall.version,DRIVER_VERSION);
  strcpy(driver_infos_overall.autor,DRIVER_AUTHOR);
  driver_infos_overall.comm_version = PROTOCOL_VERSION;
    
  discroot = NULL;
}



/*
 * Alle zu benutzenden FDs in die fdmask eintragen
 */
void update_fdmask(fd_set *fdmask)
{
  s_portinfo *ports;
  s_listen *listento;
  int i;
  
  FD_ZERO(fdmask);
  FD_SET(sockfd,fdmask);
  if (gnomfd != -1)
  {
    FD_SET(gnomfd,fdmask);
    maxfd = gnomfd+1;
  }
  
  if (monifd != -1)
     FD_SET(monifd,fdmask);
  
  ports = portsroot;
  while (ports != NULL)
  {
    FD_SET(ports->fd,fdmask);
    listento = ports->listen;
    while (listento != NULL)
    {
      FD_SET(listento->fd,fdmask);
      listento = listento->next;
    }
    ports = ports->next;
  }
  
  i = 0;
  while (i < MAX_QSOs)
  {
    if (qso[i] != NULL) FD_SET(qso[i]->fd,fdmask);
    i++;
  }
}



int main()
{
  fd_set fdmask;
  int clilen;
  struct sockaddr_un cli_addr;
  s_portinfo *ports;
  int i;
  struct timeval tv;
  s_listen *listento;
  
  
  printf("\nKernel-AX25-Treiber fr GNOM V%s\nCopyright (C) 1997 Jochen Sarrazin, DG6VJ\n\n",DRIVER_VERSION);
  
  init_globals();
  config_readconfig();
  ax25_load_info();
  ax25_open_ports();
  comm_create_and_listen_to_socket();
  config_locallist2portslist();
  monitor_init();
  
  /* Das ganze QSO-Array leer machen */
  ax25_init_qsos();
  
  /* Signalhandler installieren, damit der Daemon auch korrekt beendet werden kann */
  signal(SIGTERM,(void *)MainTerminate);
  
  
  /* Umschalten auf "daemon-mode" */
  setsid();
  umask(0);
  if (fork() != 0) _exit(0);
  
  
  
  for (;;)
  {
    update_fdmask(&fdmask);
    maxfd = getmaxfd();

    tv.tv_usec = 50000;		/* Alle 50 msec nach was zum Senden gucken. Wenn ganz schnelle
                        	 * Links langsam werden, kann dieser Wert 'runter gesetzt werden. */
    tv.tv_sec = 0;		/* Evtl. mu dieser Wert bei 9k6 oder schneller verndert werden */
    
    
    if (select(maxfd,&fdmask,NULL,NULL,&tv) > 0)
    {
      /* Befehle von GNOM? */
      if (gnomfd != -1)
         if (FD_ISSET(gnomfd,&fdmask))
            comm_check_command();
      
      /* Monitor-Daten? */
      if (monifd != -1)
         if (FD_ISSET(monifd,&fdmask))
            monitor_operate();
      
      
      /* Hat sich GNOM gemeldet? */
      if (FD_ISSET(sockfd,&fdmask))
      {
        /* Da will sich jemand anmelden */
        clilen = sizeof(cli_addr);
        gnomfd = accept(sockfd,(struct sockaddr *) &cli_addr,&clilen);
        if (gnomfd < 0)
           syslog(LOG_ERR,"Accept error (main.c)");
        else
           if (maxfd < gnomfd+1) maxfd = gnomfd+1;
      }
      else
      {
        /* Oder war es ein anderer, schon bestehender Port? */
        ports = portsroot;
        while (ports != NULL)
        {
          if (FD_ISSET(ports->fd,&fdmask))
          {
             /* Ein neuer Connect! */
             ax25_accept_qso(ports->fd,ports,ports->addr);
          }
          listento = ports->listen;
          while (listento != NULL)
          {
            if (FD_ISSET(listento->fd,&fdmask))
            {
               /* Ein neuer Connect! */
               ax25_accept_qso(listento->fd,ports,listento->call);
            }
            listento = listento->next;
          }
          ports = ports->next;
        }
      
        /* Gucken, ob von einem QSO-Port was kommt */
        i = 0;
        while (i < MAX_QSOs)
        {
          if (qso[i] != NULL)
          {
            /* Dieser Port ist benutzt */
            if (FD_ISSET(qso[i]->fd,&fdmask))
            {
              ax25_get_qso_data(i);
            }
          }
          i++;
        }
      }
    }
    else 
    {
      poll_ax25();
    }
  }
}


